#include <windows.h>
#include <gl/glu.h>
#include <gl/gl.h>
#include <math.h>

#define VK_W  0x057
#define VK_S  0x053
#define VK_A  0x041
#define VK_D  0x044

typedef struct TVector3D
{
   GLdouble  x;
   GLdouble  y;
   GLdouble  z;
}
TVector3D;

typedef struct TCameraObject
{
   TVector3D  Pos;  //   .
   TVector3D  View; //   .
   TVector3D  Up;   //  .
}
TCameraObject;

HDC hDC;
HGLRC hRC;
HWND hMainWindow;
int AppWidth, AppHeight;

TCameraObject Camera = {{0.0, 2.5, 5.0}, {0.0, 2.5, 0.0}, {0.0, 1.0, 0.0}}; //      

TVector3D NormalizeVector(TVector3D Vector)
{
   TVector3D Result;

   double Length = sqrt((Vector.x * Vector.x) + (Vector.y * Vector.y) + (Vector.z * Vector.z));

   Result.x = Vector.x / Length;
   Result.y = Vector.y / Length;
   Result.z = Vector.z / Length;

   return Result;
}

TVector3D CrossVectors(TVector3D Vector1, TVector3D Vector2)
{
   TVector3D Result;

   Result.x = ((Vector1.y * Vector2.z) - (Vector1.z * Vector2.y));
   Result.y = ((Vector1.z * Vector2.x) - (Vector1.x * Vector2.z));
   Result.z = ((Vector1.x * Vector2.y) - (Vector1.y * Vector2.x));

   return Result;
}

void RotateCamera(double Speed)
{
   TVector3D Vector;

   Vector.x = Camera.View.x - Camera.Pos.x;
   Vector.z = Camera.View.z - Camera.Pos.z;

   Camera.View.z = Camera.Pos.z + sin(Speed) * Vector.x + cos(Speed) * Vector.z;
   Camera.View.x = Camera.Pos.x + cos(Speed) * Vector.x - sin(Speed) * Vector.z;
}

void MoveCamera(double Speed)
{
   TVector3D Vector = {0};
   /*
      () 
   */
   Vector.x = Camera.View.x - Camera.Pos.x;
   Vector.z = Camera.View.z - Camera.Pos.z;

   Vector = NormalizeVector(Vector);
   /*
      ,   ,        .
          ,  Speed  - .
   */
   Camera.Pos.x  += Vector.x * Speed;
   Camera.Pos.z  += Vector.z * Speed;
   Camera.View.x += Vector.x * Speed;
   Camera.View.z += Vector.z * Speed;
}

void DrawGrid()
{
   for (float i = -50; i <= 100; i += 5)
   {
      glBegin(GL_LINES);
          glVertex3f(-50, 0, i);
          glVertex3f(100, 0, i);
          glVertex3f(i, 0,-50);
          glVertex3f(i, 0, 100);
      glEnd();
   }
}

void RenderScene(HDC hDC)
{
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); //      .
   glLoadIdentity();
   /*
         .
   */
   gluLookAt(Camera.Pos.x,  Camera.Pos.y,  Camera.Pos.z,
             Camera.View.x, Camera.View.y, Camera.View.z,
             Camera.Up.x,   Camera.Up.y,   Camera.Up.z);

   DrawGrid(); //  .
   glTranslatef(0.0f, 30.0, 0.0); //     30 -   Y)
   DrawGrid();;

   SwapBuffers(hDC);
}

void MouseMove(int xMouse, int yMouse)
{
   double xAngle = (double)((AppWidth >> 1)  - xMouse) * 0.0005;
   double yAngle = (double)((AppHeight >> 1) - yMouse) * 0.0010;

   if (Camera.View.y > 8.0)  Camera.View.y = 8.0;
   if (Camera.View.y < -3.0) Camera.View.y = -3.0;

   Camera.View.y += yAngle;
   RotateCamera(-xAngle);

   SetCursorPos(AppWidth >> 1, AppHeight >> 1);
   RenderScene(hDC);
}

void KeyboardInput(int Key)
{
   switch (Key)
   {
     case VK_W: MoveCamera(0.5);      //  .
                break;
     case VK_S: MoveCamera(-0.5);     // .
                break;
     case VK_D: RotateCamera(0.025);  // .
                break;
     case VK_A: RotateCamera(-0.025); // 
                break;

     case VK_ESCAPE:
              {
                PostQuitMessage(0);
              }
   }
}

void ResizeScene(int nWidth, int nHeight)
{
   AppWidth  = nWidth;
   AppHeight = nHeight;

   glViewport(0, 0, nWidth, nHeight);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(45.0, nWidth / nHeight, 0.1, 0); //     45 ,  ,    .
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
}

void EnableOpenGL(HWND hWindow, HDC* hDC, HGLRC* hRC)
{
   PIXELFORMATDESCRIPTOR PixelFormat = {0};

   PixelFormat.nSize      = sizeof(PIXELFORMATDESCRIPTOR);
   PixelFormat.nVersion   = 1;
   PixelFormat.dwFlags    = PFD_DRAW_TO_WINDOW |
                            PFD_SUPPORT_OPENGL |
                            PFD_DOUBLEBUFFER;
   PixelFormat.iPixelType = PFD_TYPE_RGBA;
   PixelFormat.cColorBits = 24;              //  .
   PixelFormat.cDepthBits = 16;              //   .
   PixelFormat.iLayerType = PFD_MAIN_PLANE;  //  .

   if (!(*hDC = GetDC(hWindow)))
      return;

   if (!(SetPixelFormat(*hDC, (ChoosePixelFormat(*hDC, &PixelFormat)), &PixelFormat)))
      return;

   if (!(*hRC  = wglCreateContext(*hDC)))
      return;

   if (!(wglMakeCurrent(*hDC, *hRC)))
      return;
}

LRESULT CALLBACK WindowProc(HWND hWindow, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
   switch (uMsg)
   {
      case WM_SIZE      : ResizeScene(LOWORD(lParam), (HIWORD(lParam))); //     .
                          break;
      case WM_PAINT     : RenderScene(hDC);                              //  .
                          break;
      case WM_CREATE    : EnableOpenGL(hWindow, &hDC, &hRC);             //  OpenGL.
                          break;
      case WM_KEYDOWN   : KeyboardInput(wParam);                         //     .
                          break;
      case WM_MOUSEMOVE : MouseMove(LOWORD(lParam), (HIWORD(lParam)));   //   .
                          break;
      default:
           return DefWindowProc(hWindow, uMsg, wParam, lParam);
   }
   return 0;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
   MSG Msg;
   WNDCLASSEX WindowClass;

   WindowClass.cbSize        = sizeof(WNDCLASSEX);
   WindowClass.style         = CS_OWNDC;
   WindowClass.lpfnWndProc   = WindowProc;
   WindowClass.cbClsExtra    = 0;
   WindowClass.cbWndExtra    = 0;
   WindowClass.hInstance     = hInstance;
   WindowClass.hIcon         = LoadIcon(0, IDI_APPLICATION);
   WindowClass.hCursor       = LoadCursor(0, IDC_ARROW);
   WindowClass.hbrBackground = 0;
   WindowClass.lpszMenuName  = 0;
   WindowClass.lpszClassName = "GLClass";
   WindowClass.hIconSm       = LoadIcon(0, IDI_APPLICATION);

   if (!RegisterClassEx(&WindowClass))
   {
      MessageBox(0, "   .", "!", MB_ICONERROR);
      return 0;
   }

   hMainWindow = CreateWindowEx(0,
                                "GLClass",
                                "OpenGL 3D",
                                WS_POPUPWINDOW, //     .
                                0,
                                0,
                                800,            //  .
                                600,            // .
                                0,
                                0,
                                hInstance,
                                0);
    if (!hMainWindow)
    {
      MessageBox(0, "  .", "!", MB_ICONERROR);
      return 0;
    }

    ShowWindow(hMainWindow, SW_MAXIMIZE);
    ShowCursor(0);

    while (Msg.message != WM_QUIT)
    {
       if (PeekMessage(&Msg, 0, 0, 0, PM_REMOVE))
       {
          TranslateMessage(&Msg);
          DispatchMessage(&Msg);
       }
       else
          SendMessage(hMainWindow, WM_PAINT, 0, 0); //       . (    )
    }
   return Msg.wParam;
}
